from rdflib import URIRef

# All SBOL objects will be created in the default namespace,
# unless otherwise specified
DEFAULT_NS = URIRef("http://examples.org/")

# The URIs defined here determine the appearance of serialized RDF/XML nodes.
# Change these URIs to change the appearance of an SBOL class or property name
SBOL_URI = URIRef("http://sbols.org/v2")  # < Namespace for the SBOL standard.
RDF_URI = URIRef("http://www.w3.org/1999/02/22-rdf-syntax-ns#")
PURL_URI = URIRef("http://purl.org/dc/terms/")
PROV_URI = URIRef("http://www.w3.org/ns/prov")
PROVO = URIRef("http://www.w3.org/ns/prov")
SYSBIO_URI = URIRef("http://sys-bio.org")
# PROVO = URIRef("https://www.w3.org/TR/prov-o/")

# rdf nodes used in SBOL
NODENAME_ABOUT = URIRef("rdf:about")
NODENAME_RESOURCE = URIRef("rdf:resource")

VERSION_STRING = "1"

# URIs for SBOL objects
SBOL_DOCUMENT = URIRef(SBOL_URI + "#Document")
SBOL_IDENTIFIED = URIRef(SBOL_URI + "#Identified")
SBOL_DOCUMENTED = URIRef(SBOL_URI + "#Documented")
SBOL_TOP_LEVEL = URIRef(SBOL_URI + "#TopLevel")
SBOL_GENERIC_TOP_LEVEL = URIRef(SBOL_URI + "#GenericTopLevel")
SBOL_SEQUENCE_ANNOTATION = URIRef(SBOL_URI + "#SequenceAnnotation")
SBOL_COMPONENT = URIRef(SBOL_URI + "#Component")
SBOL_FUNCTIONAL_COMPONENT = URIRef(SBOL_URI + "#FunctionalComponent")
SBOL_COMPONENT_DEFINITION = URIRef(SBOL_URI + "#ComponentDefinition")
SBOL_SEQUENCE = URIRef(SBOL_URI + "#Sequence")
SBOL_MODULE_DEFINITION = URIRef(SBOL_URI + "#ModuleDefinition")
SBOL_MODULE = URIRef(SBOL_URI + "#Module")
SBOL_MODEL = URIRef(SBOL_URI + "#Model")
SBOL_MAPS_TO = URIRef(SBOL_URI + "#MapsTo")
SBOL_INTERACTION = URIRef(SBOL_URI + "#Interaction")
SBOL_PARTICIPATION = URIRef(SBOL_URI + "#Participation")
SBOL_SEQUENCE_CONSTRAINT = URIRef(SBOL_URI + "#SequenceConstraint")
SBOL_LOCATION = URIRef(SBOL_URI + "#Location")
# SBOL_DOCUMENT = URIRef(SBOL_URI + "#Document")
SBOL_RANGE = URIRef(SBOL_URI + "#Range")
SBOL_CUT = URIRef(SBOL_URI + "#Cut")
SBOL_COLLECTION = URIRef(SBOL_URI + "#Collection")
SBOL_GENERIC_LOCATION = URIRef(SBOL_URI + "#GenericLocation")
SBOL_VARIABLE_COMPONENT = URIRef(SBOL_URI + "#VariableComponent")
SBOL_COMBINATORIAL_DERIVATION = URIRef(SBOL_URI + "#CombinatorialDerivation")
SBOL_ATTACHMENT = URIRef(SBOL_URI + "#Attachment")
SBOL_IMPLEMENTATION = URIRef(SBOL_URI + "#Implementation")
SBOL_EXPERIMENT = URIRef(SBOL_URI + "#Experiment")
SBOL_EXPERIMENTAL_DATA = URIRef(SBOL_URI + "#ExperimentalData")
SBOL_MEASURE = URIRef(SBOL_URI + "#Measure")
UNDEFINED = URIRef(SBOL_URI + "#Undefined")

# URIs for SBOL Properties
SBOL_IDENTITY = URIRef(SBOL_URI + "#identity")
SBOL_PERSISTENT_IDENTITY = URIRef(SBOL_URI + "#persistentIdentity")
SBOL_VERSION = URIRef(SBOL_URI + "#version")
SBOL_DISPLAY_ID = URIRef(SBOL_URI + "#displayId")
SBOL_NAME = URIRef(PURL_URI + "title")
SBOL_DESCRIPTION = URIRef(PURL_URI + "description")
SBOL_TYPES = URIRef(SBOL_URI + "#type")
SBOL_START = URIRef(SBOL_URI + "#start")
SBOL_END = URIRef(SBOL_URI + "#end")
SBOL_SEQUENCE_ANNOTATIONS = URIRef(SBOL_URI + "#sequenceAnnotation")
SBOL_COMPONENTS = URIRef(SBOL_URI + "#component")
SBOL_COMPONENT_PROPERTY = URIRef(SBOL_URI + "#component")
SBOL_ROLES = URIRef(SBOL_URI + "#role")
SBOL_ELEMENTS = URIRef(SBOL_URI + "#elements")
SBOL_ENCODING = URIRef(SBOL_URI + "#encoding")
SBOL_SEQUENCE_PROPERTY = URIRef(SBOL_URI + "#sequence")
SBOL_WAS_DERIVED_FROM = URIRef(PROV_URI + "#wasDerivedFrom")
SBOL_DEFINITION = URIRef(SBOL_URI + "#definition")
SBOL_ACCESS = URIRef(SBOL_URI + "#access")
SBOL_DIRECTION = URIRef(SBOL_URI + "#direction")
SBOL_MODELS = URIRef(SBOL_URI + "#model")
SBOL_MODULES = URIRef(SBOL_URI + "#module")
SBOL_FUNCTIONAL_COMPONENTS = URIRef(SBOL_URI + "#functionalComponent")
SBOL_INTERACTIONS = URIRef(SBOL_URI + "#interaction")
SBOL_MAPS_TOS = URIRef(SBOL_URI + "#mapsTo")
SBOL_PARTICIPATIONS = URIRef(SBOL_URI + "#participation")
SBOL_PARTICIPANT = URIRef(SBOL_URI + "#participant")
SBOL_LOCAL = URIRef(SBOL_URI + "#local")
SBOL_REMOTE = URIRef(SBOL_URI + "#remote")
SBOL_REFINEMENT = URIRef(SBOL_URI + "#refinement")
SBOL_SOURCE = URIRef(SBOL_URI + "#source")
SBOL_LANGUAGE = URIRef(SBOL_URI + "#language")
SBOL_FRAMEWORK = URIRef(SBOL_URI + "#framework")
SBOL_SEQUENCE_CONSTRAINTS = URIRef(SBOL_URI + "#sequenceConstraint")
SBOL_SUBJECT = URIRef(SBOL_URI + "#subject")
SBOL_OBJECT = URIRef(SBOL_URI + "#object")
SBOL_RESTRICTION = URIRef(SBOL_URI + "#restriction")
SBOL_ORIENTATION = URIRef(SBOL_URI + "#orientation")
SBOL_LOCATIONS = URIRef(SBOL_URI + "#location")
SBOL_ROLE_INTEGRATION = URIRef(SBOL_URI + "#roleIntegration")
SBOL_MEMBERS = URIRef(SBOL_URI + "#member")
SBOL_AT = URIRef(SBOL_URI + "#at")
SBOL_OPERATOR = URIRef(SBOL_URI + "#operator")
SBOL_VARIABLE_COMPONENTS = URIRef(SBOL_URI + "#variableComponent")
SBOL_VARIABLE = URIRef(SBOL_URI + "#variable")
SBOL_VARIANTS = URIRef(SBOL_URI + "#variants")
SBOL_VARIANT_COLLECTIONS = URIRef(SBOL_URI + "#variantCollections")
SBOL_VARIANT_DERIVATIONS = URIRef(SBOL_URI + "#variantDeriviations")
SBOL_STRATEGY = URIRef(SBOL_URI + "#strategy")
SBOL_TEMPLATE = URIRef(SBOL_URI + "#template")
SBOL_ATTACHMENTS = URIRef(SBOL_URI + "#attachment")
SBOL_MEASUREMENTS = URIRef(SBOL_URI + "#measure")
SBOL_UNIT = URIRef(SBOL_URI + "#unit")
SBOL_VALUE = URIRef(SBOL_URI + "#hasNumericalValue")

# SBOL internal ontologies
# Option for Component::access or FunctionalComponent::access property
SBOL_ACCESS_PRIVATE = URIRef(SBOL_URI + "#private")
# Option for Component::access or FunctionalComponent::access property
SBOL_ACCESS_PUBLIC = URIRef(SBOL_URI + "#public")
# Option for FunctionalComponent::access property
SBOL_DIRECTION_IN = URIRef(SBOL_URI + "#in")
# Option for FunctionalComponent::direction property
SBOL_DIRECTION_OUT = URIRef(SBOL_URI + "#out")
# Option for FunctionalComponent::direction property
SBOL_DIRECTION_IN_OUT = URIRef(SBOL_URI + "#inout")
# Option for FunctionalComponent::direction property
SBOL_DIRECTION_NONE = URIRef(SBOL_URI + "#none")
# Option for SequenceConstraint::restriction property
SBOL_RESTRICTION_PRECEDES = URIRef(SBOL_URI + "#precedes")
# Option for SequenceConstraint::restriction property
SBOL_RESTRICTION_SAME_ORIENTATION_AS = URIRef("#sameOrientationAs")
# Option for SequenceConstraint::restriction property
SBOL_RESTRICTION_OPPOSITE_ORIENTATION_AS = URIRef("#oppositeOrientationAs")
# Option for Sequence::encoding property
SBOL_ENCODING_IUPAC = URIRef("http://www.chem.qmul.ac.uk/iubmb/misc/naseq.html")
# Option for Sequence::encoding property
SBOL_ENCODING_IUPAC_PROTEIN = URIRef("http://www.chem.qmul.ac.uk/iupac/AminoAcid/")
# Option for Sequence::encoding property
SBOL_ENCODING_SMILES = URIRef("http://www.opensmiles.org/opensmiles.html")
# Option for Location::orientation property
SBOL_ORIENTATION_INLINE = URIRef(SBOL_URI + "#inline")
# Option for Location::orientation property
SBOL_ORIENTATION_REVERSE_COMPLEMENT = URIRef(SBOL_URI + "#reverseComplement")
# Option for FunctionalComponent::refinement property
SBOL_REFINEMENT_USE_REMOTE = URIRef(SBOL_URI + "#useRemote")
# Option for FunctionalComponent::refinement property
SBOL_REFINEMENT_USE_LOCAL = URIRef(SBOL_URI + "#useLocal")
# Option for MapsTo::refinement property
SBOL_REFINEMENT_VERIFY_IDENTICAL = URIRef(SBOL_URI + "#verifyIdentical")
# Option for MapsTo::refinement property
SBOL_REFINEMENT_MERGE = URIRef(SBOL_URI + "#merge")
# Option for SequenceAnnotation::roleIntegration
# or Component::roleIntegration property
SBOL_ROLE_INTEGRATION_MERGE = URIRef(SBOL_URI + "#mergeRoles")
# Option for SequenceAnnotation::roleIntegration
# or Component::roleIntegration property
SBOL_ROLE_INTEGRATION_OVERRIDE = URIRef(SBOL_URI + "#overrideRoles")
# Option for Usage::roles or Association::roles
SBOL_DESIGN = URIRef(SBOL_URI + "#design")
# Option for Usage::roles or Association::roles
SBOL_BUILD = URIRef(SBOL_URI + "#build")
# Option for Usage::roles or Association::roles
SBOL_TEST = URIRef(SBOL_URI + "#test")
# Option for Usage::roles or Association::roles
SBOL_LEARN = URIRef(SBOL_URI + "#learn")

# PROVO ontology
PROVO_ACTIVITY = URIRef(PROVO + "#Activity")
PROVO_USAGE = URIRef(PROVO + "#Usage")
PROVO_ASSOCIATION = URIRef(PROVO + "#Association")
PROVO_AGENT = URIRef(PROVO + "#Agent")
PROVO_PLAN = URIRef(PROVO + "#Plan")
PROVO_WAS_GENERATED_BY = URIRef(PROVO + "#wasGeneratedBy")
PROVO_STARTED_AT_TIME = URIRef(PROVO + "#startedAtTime")
PROVO_ENDED_AT_TIME = URIRef(PROVO + "#endedAtTime")
PROVO_QUALIFIED_ASSOCIATION = URIRef(PROVO + "#qualifiedAssociation")
PROVO_QUALIFIED_USAGE = URIRef(PROVO + "#qualifiedUsage")
PROVO_WAS_INFORMED_BY = URIRef(PROVO + "#wasInformedBy")
PROVO_HAD_PLAN = URIRef(PROVO + "#hadPlan")
PROVO_HAD_ROLE = URIRef(PROVO + "#hadRole")
PROVO_AGENT_PROPERTY = URIRef(PROVO + "#agent")
PROVO_ENTITY = URIRef(PROVO + "#entity")

# Systems Biology Ontology
# Interaction.types
# Namespace for Systems Biology Ontology (SBO) terms
SBO = URIRef("http://identifiers.org/biomodels.sbo/SBO:")
# An SBO term and possible value for an Interaction::type property
SBO_INTERACTION = URIRef(SBO + "0000343")
# An SBO term and possible value for an Interaction::type property
SBO_INHIBITION = URIRef(SBO + "0000169")
# An SBO term and possible value for an Interaction::type property
SBO_GENETIC_PRODUCTION = URIRef(SBO + "0000589")
# An SBO term and possible value for an Interaction::type property
SBO_NONCOVALENT_BINDING = URIRef(SBO + "0000177")
# An SBO term and possible value for an Interaction::type property
SBO_STIMULATION = URIRef(SBO + "0000170")
# An SBO term and possible value for an Interaction::type property
SBO_DEGRADATION = URIRef(SBO + "0000179")
# An SBO term and possible value for an Interaction::type property
SBO_CONTROL = URIRef(SBO + "0000168")
# An SBO term and possible value for an Interaction::type property
SBO_BIOCHEMICAL_REACTION = URIRef(SBO + "0000176")
# An SBO term and possible value for an Interaction::type property
SBO_STIMULATED = URIRef(SBO + "0000643")
# An SBO term and possible value for an Interaction::type property
SBO_CONVERSION = URIRef(SBO + "0000182")

# Participant.roles
# An SBO term and possible value for an Participant::role property
SBO_PROMOTER = URIRef(SBO + "0000598")
# An SBO term and possible value for an Participant::role property
SBO_GENE = URIRef(SBO + "0000243")
# An SBO term and possible value for an Participant::role property
SBO_INHIBITOR = URIRef(SBO + "0000020")
# An SBO term and possible value for an Participant::role property
SBO_INHIBITED = URIRef(SBO + "0000642")
# An SBO term and possible value for an Participant::role property
SBO_STIMULATOR = URIRef(SBO + "0000459")
# An SBO term and possible value for an Participant::role property
SBO_REACTANT = URIRef(SBO + "0000010")
# An SBO term and possible value for an Participant::role property
SBO_PRODUCT = URIRef(SBO + "0000011")
# An SBO term and possible value for an Participant::role property
SBO_LIGAND = URIRef(SBO + "0000280")
# An SBO term and possible value for an Participant::role property
SBO_NONCOVALENT_COMPLEX = URIRef(SBO + "0000253")
# An SBO term and possible value for an Participant::role property
SBO_BINDING_SITE = URIRef(SBO + "0000494")
SBO_SUBSTRATE = URIRef(SBO + "0000015")
SBO_COFACTOR = URIRef(SBO + "0000604")
SBO_SIDEPRODUCT = URIRef(SBO + "0000603")
# SBO_PRODUCT = URIRef(SBO + "0000011")
SBO_ENZYME = URIRef(SBO + "0000014")

# URIs for common Sequence Ontology terms
# Namespace for Sequence Ontology (SO) terms
SO = URIRef("http://identifiers.org/so/SO:")
# An SO term and possible value for ComponentDefinition::role property
SO_MISC = URIRef(SO + "0000001")
# An SO term and possible value for ComponentDefinition::role property
SO_GENE = URIRef(SO + "0000704")
# An SO term and possible value for ComponentDefinition::role property
SO_PROMOTER = URIRef(SO + "0000167")
# An SO term and possible value for ComponentDefinition::role property
SO_CDS = URIRef(SO + "0000316")
# An SO term and possible value for ComponentDefinition::role property
SO_RBS = URIRef(SO + "0000139")
# An SO term and possible value for ComponentDefinition::role property
SO_TERMINATOR = URIRef(SO + "0000141")
# An SO term and possible value for ComponentDefinition::role property
SO_SGRNA = URIRef(SO + "0001998")
# An SO term and possible value for ComponentDefinition::role property
SO_LINEAR = URIRef(SO + "0000987")
# An SO term and possible value for ComponentDefinition::role property
SO_CIRCULAR = URIRef(SO + "0000988")
# An SO term and possible value for ComponentDefinition::role property
SO_PLASMID = URIRef(SO + "0000155")

# BioPAX is used to indicate macromolecular and molecular types
# DNA
# A BioPax term and possible value for ComponentDefinition::type property
BIOPAX_DNA = URIRef("http://www.biopax.org/release/biopax-level3.owl#DnaRegion")
# RNA
# A BioPax term and possible value for ComponentDefinition::type property
BIOPAX_RNA = URIRef("http://www.biopax.org/release/biopax-level3.owl#RnaRegion")
# PROTEIN
# A BioPax term and possible value for ComponentDefinition::type property
BIOPAX_PROTEIN = URIRef("http://www.biopax.org/release/biopax-level3.owl#Protein")
# SMALL_MOLECULE
# A BioPax term and possible value for ComponentDefinition::type property
BIOPAX_SMALL_MOLECULE = URIRef("http://www.biopax.org/release/biopax-level3.owl#SmallMolecule")
# COMPLEX
# A BioPax term and possible value for ComponentDefinition::type property
BIOPAX_COMPLEX = URIRef("http://www.biopax.org/release/biopax-level3.owl#Complex")

# EDAM ontology is used for Model.languages
# An EDAM ontology term and option for Model::language
EDAM_SBML = URIRef("http://identifiers.org/edam/format_2585")
# An EDAM ontology term and option for Model::language
EDAM_CELLML = URIRef("http://identifiers.org/edam/format_3240")
# An EDAM ontology term and option for Model::language
EDAM_BIOPAX = URIRef("http://identifiers.org/edam/format_3156")

# Model.frameworks
# SBO term and option for Model::framework
SBO_CONTINUOUS = URIRef(SBO + "0000062")
# SBO term and option for Model::framework
SBO_DISCRETE = URIRef(SBO + "0000063")

# URIs for SBOL extension objects
SYSBIO_DESIGN = URIRef(SYSBIO_URI + "#Design")
SYSBIO_BUILD = URIRef(SYSBIO_URI + "#Build")
SYSBIO_TEST = URIRef(SYSBIO_URI + "#Test")
SYSBIO_ANALYSIS = URIRef(SYSBIO_URI + "#Analysis")
SYSBIO_SAMPLE_ROSTER = URIRef(SYSBIO_URI + "#SampleRoster")

IGEM_URI = URIRef("http://wiki.synbiohub.org/wiki/Terms/igem")
IGEM_STANDARD_ASSEMBLY = URIRef(IGEM_URI + "#assembly/RFC10")



from rdflib import URIRef

from .constants import *
from .identified import Identified
from .measurement import Measurement
from .property import URIProperty, ReferencedObject, OwnedObject


class Participation(Identified):

    def __init__(self, type_uri=SBOL_PARTICIPATION, uri=URIRef('example'),
                 participant='', version=VERSION_STRING):
        super().__init__(type_uri, uri, version)
        self._roles = URIProperty(self, SBOL_ROLES, '0', '*', [])
        self.participant = ReferencedObject(self, SBOL_PARTICIPANT,
                                            SBOL_FUNCTIONAL_COMPONENT,
                                            '1', '1', [], participant)
        self.measurements = OwnedObject(self, SBOL_MEASUREMENTS,
                                        Measurement, '0', '*', [])

    @property
    def roles(self):
        return self._roles.value

    @roles.setter
    def roles(self, new_roles):
        self._roles.set(new_roles)

    def addRole(self, new_role):
        self._roles.add(new_role)

    def removeRole(self, index=0):
        self._roles.remove(index)

    def define(self, species, role=""):
        raise NotImplementedError('Not yet implemented')



from .constants import *
from .identified import Identified
from .location import Location
from .mapsto import MapsTo
from .measurement import Measurement
from .property import URIProperty, OwnedObject, ReferencedObject


class ComponentInstance(Identified):
    def __init__(self, rdf_type, uri, definition, access, version):
        super().__init__(rdf_type, uri, version)
        self.definition = ReferencedObject(self, SBOL_DEFINITION,
                                           SBOL_COMPONENT_DEFINITION,
                                           '1', '1', [], definition)
        self._access = URIProperty(self, SBOL_ACCESS, '0', '1', [], access)
        self.mapsTos = OwnedObject(self, SBOL_MAPS_TOS, MapsTo,
                                   '0', '*', [])
        self.measurements = OwnedObject(self, SBOL_MEASUREMENTS, Measurement,
                                        '0', '*', [])

    @property
    def access(self):
        return self._access.value

    @access.setter
    def access(self, new_access):
        self._access.set(new_access)


class Component(ComponentInstance):
    def __init__(self, uri=URIRef('example'), definition='',
                 access=SBOL_ACCESS_PUBLIC, version=VERSION_STRING):
        super().__init__(SBOL_COMPONENT, uri, definition, access, version)
        self._roles = URIProperty(self, SBOL_ROLES, '0', '*', [])
        self._roleIntegration = URIProperty(self, SBOL_ROLE_INTEGRATION,
                                            '0', '1', [],
                                            SBOL_ROLE_INTEGRATION_MERGE)
        self.sourceLocations = OwnedObject(self, SBOL_LOCATIONS, Location,
                                           '0', '*', [])

    @property
    def roles(self):
        return self._roles.value

    @roles.setter
    def roles(self, new_roles):
        self._roles.set(new_roles)

    def addRole(self, new_role):
        self._roles.add(new_role)

    def removeRole(self, index=0):
        self._roles.remove(index)

    @property
    def roleIntegration(self):
        return self._roleIntegration.value

    @roleIntegration.setter
    def roleIntegration(self, new_roleIntegration):
        self._roleIntegration.set(new_roleIntegration)


class FunctionalComponent(ComponentInstance):
    def __init__(self, uri=URIRef('example'), definition='',
                 access=SBOL_ACCESS_PUBLIC, direction=SBOL_DIRECTION_NONE,
                 version=VERSION_STRING):
        super().__init__(SBOL_FUNCTIONAL_COMPONENT, uri,
                         definition, access, version)
        self._direction = URIProperty(self, SBOL_DIRECTION,
                                      '1', '1', [], direction)

    @property
    def direction(self):
        return self._direction.value

    @direction.setter
    def direction(self, new_direction):
        self._direction.set(new_direction)

    def connect(self, interface_component):
        raise NotImplementedError("Not yet implemented")

    def mask(self, masked_component):
        raise NotImplementedError("Not yet implemented")

    def override(selfself, masked_component):
        raise NotImplementedError("Not yet implemented")

    def isMasked(self):
        raise NotImplementedError("Not yet implemented")



from .toplevel import TopLevel
from .constants import *
from .property import URIProperty


class Implementation(TopLevel):

    def __init__(self, uri=URIRef("example"), version=VERSION_STRING,
                 *, type_uri=SBOL_IMPLEMENTATION):
        super().__init__(type_uri, uri, version)
        self._built = URIProperty(self, SBOL_URI+'#built', '0', '1', [])

    @property
    def built(self):
        return self._built.value

    @built.setter
    def built(self, new_built):
        self._built.set(new_built)



import posixpath

from .object import SBOLObject
from .config import Config
from .config import ConfigOptions
from .config import getHomespace
from .config import hasHomespace
from .constants import *
from .property import LiteralProperty
from .property import ReferencedObject
from .property import URIProperty
from .sbolerror import SBOLError
from .sbolerror import SBOLErrorCode
from .config import parseClassName
from . import validation


class Identified(SBOLObject):
    # The persistentIdentity property is OPTIONAL and has a data type of URI.
    # This URI serves to uniquely refer to a set of SBOL objects
    # that are different versions of each other.
    # An Identified object MUST be referred to
    # using either its identity URI or its persistentIdentity URI.
    # _persistentIdentity = None

    # The displayId property is an OPTIONAL identifier
    # with a data type of String.
    # This property is intended to be an intermediate
    # between name and identity that is machine-readable,
    # but more human-readable than the full URI
    # of an identity. If the displayId property is used,
    # then its String value SHOULD be locally unique
    # (global uniqueness is not necessary)
    # and MUST be composed of only alphanumeric or underscore characters
    # and MUST NOT begin with a digit.
    # _displayId = None

    # If the version property is used,
    # then it is RECOMMENDED that version numbering follow the conventions
    # of [semantic versioning](http://semver.org/),
    # particularly as implemented by [Maven](http://maven.apache.org/).
    # This convention represents versions as sequences
    # of numbers and qualifiers that are separated
    # by the characters '.' and '-' and are compared in lexicographical order
    # (for example, 1 < 1.3.1 < 2.0-beta).
    # For a full explanation, see the linked resources.
    # _version = None

    # The wasDerivedFrom property is OPTIONAL and has a data type of URI.
    # An SBOL object with this property refers to another SBOL object
    # or non-SBOL resource from which this object was derived.
    # If the wasDerivedFrom property of an SBOL object A
    # that refers to an SBOL object B has an identical persistentIdentity,
    # and both A and B have a version,
    # then the version of B MUST precede that of A.
    # In addition, an SBOL object MUST NOT refer to itself
    # via its own wasDerivedFrom property
    # or form a cyclical chain of references via its wasDerivedFrom property
    # and those of other SBOL objects. For example, the reference chain
    # "A was derived from B and B was derived from A" is cyclical.
    # _wasDerivedFrom = None

    # An Activity which generated this ComponentDefinition,
    # eg., a design process like codon-optimization
    # or a construction process like Gibson Assembly
    # _wasGeneratedBy = None

    # The name property is OPTIONAL and has a data type of
    # String. This property is intended to be displayed to a human
    # when visualizing an Identified object. If an Identified object
    # lacks a name, then software tools SHOULD instead display the
    # object's displayId or identity. It is RECOMMENDED that software
    # tools give users the ability to switch perspectives between name
    # properties that are human-readable and displayId properties that
    # are less human-readable, but are more likely to be unique.
    # _name = None

    # The description property is OPTIONAL and has a data type of String.
    # This property is intended to contain a more thorough
    # text description of an Identified object.
    # _description = None

    def __init__(self, type_uri=SBOL_IDENTIFIED, uri=URIRef('example'),
                 version=VERSION_STRING):
        super().__init__(type_uri, uri)
        self._persistentIdentity = URIProperty(self, SBOL_PERSISTENT_IDENTITY,
                                               '0', '1', None, URIRef(uri))
        self._displayId = LiteralProperty(self, SBOL_DISPLAY_ID, '0', '1',
                                          [validation.sbol_rule_10204])
        self._version = LiteralProperty(self, SBOL_VERSION, '0', '1', None, version)
        self._name = LiteralProperty(self, SBOL_NAME, '0', '1', None)
        self._description = LiteralProperty(self, SBOL_DESCRIPTION, '0', '1', None)
        if Config.getOption(ConfigOptions.SBOL_COMPLIANT_URIS.value) is True:
            self._displayId.set(uri)
            self._persistentIdentity.set(URIRef(posixpath.join(getHomespace(), uri)))
            if Config.getOption(ConfigOptions.SBOL_TYPED_URIS.value) is True:
                if self.version:
                    self._identity.set(
                        URIRef(posixpath.join(getHomespace(),
                                              self.getClassName(type_uri),
                                              uri, self.version))
                    )
                else:
                    self._identity.set(
                        URIRef(posixpath.join(getHomespace(),
                                              self.getClassName(type_uri),
                                              uri))
                    )
            else:
                if self.version:
                    self._identity.set(
                        URIRef(posixpath.join(getHomespace(), uri, self.version)))
                else:
                    self._identity.set(
                        URIRef(posixpath.join(getHomespace(), uri)))
        elif hasHomespace():
            self._identity.set(URIRef(posixpath.join(getHomespace(), uri)))
            self._persistentIdentity.set(
                URIRef(posixpath.join(getHomespace(), uri)))
        # Provo hooks
        self._wasDerivedFrom = URIProperty(self, SBOL_WAS_DERIVED_FROM,
                                           '0', '*', None)
        self.wasGeneratedBy = ReferencedObject(self, PROVO_WAS_GENERATED_BY,
                                               PROVO_WAS_GENERATED_BY, '0', '*', [])
        # self._identity.validate() # TODO

    @property
    def persistentIdentity(self):
        return self._persistentIdentity.value

    @persistentIdentity.setter
    def persistentIdentity(self, new_persistentIdentity):
        self._persistentIdentity.set(new_persistentIdentity)

    @property
    def displayId(self):
        return self._displayId.value

    @displayId.setter
    def displayId(self, new_displayId):
        self._displayId.set(new_displayId)

    @property
    def version(self):
        return self._version.value

    @version.setter
    def version(self, new_version):
        self._version.set(new_version)

    @property
    def description(self):
        return self._description.value

    @description.setter
    def description(self, new_description):
        self._description.set(new_description)

    @property
    def name(self):
        return self._name.value

    @name.setter
    def name(self, new_name):
        self._name.set(new_name)

    @property
    def wasDerivedFrom(self):
        return self._wasDerivedFrom.value

    @wasDerivedFrom.setter
    def wasDerivedFrom(self, new_wasDerivedFrom):
        self._wasDerivedFrom.set(new_wasDerivedFrom)

    def generate(self):
        raise NotImplementedError("Not yet implemented")

    def update_uri(self):
        """
        Recursively generates SBOL compliant ids for an object and all
        its owned objects, then checks to make sure that these ids are unique.
        :return: None
        """
        if self.parent is None:
            raise Exception('update_uri: Parent cannot be None')
        parent = self.parent
        if Config.getOption(ConfigOptions.SBOL_COMPLIANT_URIS.value) is True:
            # Form compliant URI for child object
            persistent_id = parent.properties[SBOL_PERSISTENT_IDENTITY][0]
            persistent_id = posixpath.join(persistent_id, self.displayId)
            if len(parent.properties[SBOL_VERSION]) > 0:
                version = parent.properties[SBOL_VERSION][0]
            else:
                version = VERSION_STRING
            obj_id = posixpath.join(persistent_id, version)
            # Reset SBOLCompliant properties
            self._identity.set(obj_id)
            self._persistentIdentity.set(persistent_id)
            # Check for uniqueness of URI in local object properties
            matches = parent.find_property_value(SBOL_IDENTIFIED, obj_id)
            if len(matches) > 0:
                raise SBOLError("Cannot update SBOL-compliant URI. The URI " +
                                str(self.identity) + " is not unique",
                                SBOLErrorCode.SBOL_ERROR_URI_NOT_UNIQUE)
            for rdf_type, store in self.owned_objects.items():
                if rdf_type not in self._hidden_properties:
                    for nested_obj in store:
                        nested_obj.update_uri()
        # Check for uniqueness of URI in Document
        if parent.doc:
            matches = parent.doc.find_property_value(SBOL_IDENTITY, self.identity)
            if len(matches) > 0:
                raise SBOLError("Cannot update SBOL-compliant URI. "
                                "An object with URI " + str(self.identity) +
                                " is already in the Document",
                                SBOLErrorCode.SBOL_ERROR_URI_NOT_UNIQUE)

    def copy(self, target_doc=None, target_namespace=None, version=None):

        new_obj = self.__class__()

        # This namespace map will be used later when copying namespaces over to
        # the new Document
        if self.doc:
            namespace_map = {ns: p for p, ns in self.doc._namespaces.items()}

        # Copy properties
        for property_uri, value_store in self.properties.items():
            new_obj.properties[property_uri] = value_store.copy()

            # Add a non-default namespace to the target document if not present
            # (This can happen when copying extension properties not in the
            # SBOL namespace, for example.)
            if self.doc and target_doc is not None:
                property_namespace = URIRef(parseNamespace(property_uri))
                if property_namespace in namespace_map.keys():
                    prefix = namespace_map[property_namespace]
                    target_doc.addNamespace(property_namespace, prefix)

        # If caller specified a target_namespace argument, then import objects into this
        # new namespace. This involves replacing the target_namespace in ReferenceObject
        # URIs with the current Homespace. Don't overwrite namespaces for the
        # wasDerivedFrom field, which points back to the original object
        if target_namespace:

            # Map the identity of self into the target namespace
            if '_identity' in self.__dict__.keys():
                old_uri = self.__dict__['_identity'].value
                new_uri = replace_namespace(old_uri, target_namespace, self.getTypeURI())
                new_obj.__dict__['_identity'].value = new_uri

            if '_persistentIdentity' in self.__dict__.keys():
                old_uri = self.__dict__['_persistentIdentity'].value
                new_uri = replace_namespace(old_uri, target_namespace, self.getTypeURI())
                new_obj.__dict__['_persistentIdentity'].value = new_uri

            # Map any references to other SBOL objects in the Document into the new
            # namespace
            if self.doc is not None:

                # Collect ReferencedObject attributes
                reference_properties = [p for p in new_obj.__dict__.values() if
                                        isinstance(p, ReferencedObject)]

                # These URIProperty attributes should be treated like ReferencedObject
                # attributes
                if '_built' in new_obj.__dict__.keys():
                    reference_properties.append(new_obj.__dict__['_built'])

                for reference_property in reference_properties:
                    values = new_obj.properties[reference_property._rdf_type]
                    new_values = []
                    for uri in values:
                        if target_namespace in uri:

                            referenced_object = self.doc.find(uri)
                            if referenced_object is None:
                                continue
                            new_uri = replace_namespace(uri, target_namespace,
                                                        referenced_object.getTypeURI())
                            new_values.append(new_uri)
                    new_obj.properties[reference_property._rdf_type] = new_values

        # Assign the new object to the target Document
        if target_doc:
            target_doc.add(new_obj)
        elif self.doc:
            self.doc.add(new_obj)

        # Set the new object's version according to the user specified parameter. If
        # user didnt't provide a version, then set it automatically based on self's
        # version (if it has one).
        if version:
            new_obj.version = version
        elif self.version is not None:

            # If user is copying into the same Document and namespace, then assume the
            # user is trying to create a new version of the object. The version should
            # be automatically incremented to avoid a URI collision with the original
            # object.  However, if user is copying into a different Document, then copy
            # the original object's version without incrementing
            if new_obj.doc and new_obj.doc is self.doc and not target_namespace:
                new_obj.version.incrementMajor()
            else:
                new_obj.version = self.version

        # When an object is simply being cloned, the value of wasDerivedFrom should be
        # copied exactly as is from self. However, when copy is being used to generate
        # a new entity, the wasDerivedFrom should point back to self.
        if self.identity == new_obj.identity:
            new_obj.wasDerivedFrom = self.wasDerivedFrom
        else:
            new_obj.wasDerivedFrom = self.identity

        # Copy child objects recursively
        for property_uri, object_list in self.owned_objects.items():
            # Don't copy hidden properties
            if target_doc and property_uri in self._hidden_properties:
                continue
            for o in object_list:
                o_copy = o.copy(target_doc, target_namespace, version)
                new_obj.owned_objects[property_uri].append(o_copy)
                o_copy.parent = self
                # o_copy.update_uri()

        return new_obj


def parseNamespace(uri):
    '''
    Utility function for parsing the namespace from an RDF type -- note, this
    will work on some SBOL object URIs, but not universally.  It is primarily
    for use with RDF types.
    '''
    rlimit = uri.rfind('#') + 1
    if rlimit:
        return uri[:rlimit]
    rlimit = uri.rfind('/') + 1
    if rlimit:
        return uri[:rlimit]
    return ''


def replace_namespace(old_uri, target_namespace, rdf_type):
    '''
    Utility function for mapping an SBOL object's identity into a new namespace. The
    rdf_type is used to map to and from sbol-typed namespaces.
    '''

    # If the value is an SBOL-typed URI, replace both the namespace and class name
    class_name = parseClassName(rdf_type)
    replacement_target = target_namespace + '/' + class_name

    # If not an sbol typed URI, then just replace the namespace
    if replacement_target not in old_uri:
        replacement_target = target_namespace

    if Config.getOption('sbol_typed_uris'):
        # Map into a typed namespace
        replacement = getHomespace() + '/' + class_name
    else:
        # Map into a non-typed namespace
        replacement = getHomespace()

    new_uri = old_uri.replace(replacement_target, replacement)
    if type(old_uri) is URIRef:
        return URIRef(new_uri)
    return new_uri



from .constants import *
from .identified import Identified
from .property import LiteralProperty, URIProperty

from rdflib import URIRef


class Measurement(Identified):
    """The purpose of the Measure class is to link a numerical value
    to a unit of measure."""

    def __init__(self, uri=URIRef('example'), value=0.0,
                 unit='', version=VERSION_STRING):
        super().__init__(SBOL_MEASURE, uri, version)
        self._value = LiteralProperty(self, SBOL_VALUE, '1', '1', [], value)
        self._unit = URIProperty(self, SBOL_UNIT, '1', '1', [], unit)
        self._types = URIProperty(self, SBOL_TYPES, '0', '*', [])

    @property
    def value(self):
        return self._value.value

    @value.setter
    def value(self, new_value):
        self._value.set(new_value)

    @property
    def unit(self):
        return self._unit.value

    @unit.setter
    def unit(self, new_unit):
        self._unit.set(new_unit)

    @property
    def types(self):
        return self._types.value

    @types.setter
    def types(self, new_types):
        self._types.set(new_types)

    def addType(self, new_type):
        self._types.add(new_type)

    def removeType(self, index=0):
        self._types.remove(index)



from deprecated import deprecated

from rdflib import URIRef

from .constants import *
from .property import LiteralProperty
from .toplevel import TopLevel


class Sequence(TopLevel):
    """
    The primary structure (eg, nucleotide or amino acid sequence)
    of a ComponentDefinition object.
    """

    def __init__(self, uri=URIRef("example"), elements="",
                 encoding=SBOL_ENCODING_IUPAC,
                 version=VERSION_STRING, type_uri=SBOL_SEQUENCE):
        """Construct a Sequence.

        :param uri: A full URI including a scheme, namespace, and identifier.
        If SBOLCompliance configuration is enabled, then this argument is
        simply the displayId for the new object and a full URI will
        automatically be constructed.
        :param elements: A string representation of the primary structure
        of DNA, RNA, protein, or a SMILES string for small molecules.
        :param encoding: A URI that describes the representation format
        used for the elements property. Set to SBOL_ENCODING_IUPAC by default
        :param version: An arbitrary version string. If SBOLCompliance is
        enabled, this should be a Maven version string.
        :param type_uri: The RDF type for an extension class derived
        from this one.
        """
        super().__init__(type_uri, uri, version)
        self._elements = LiteralProperty(self, SBOL_ELEMENTS,
                                         '1', '1', [], elements)
        self._encoding = LiteralProperty(self, SBOL_ENCODING,
                                         '1', '1', [], encoding)

    # The elements property is a REQUIRED String of characters that represents
    # the constituents of a biological or chemical molecule. For example,
    # these characters could represent the nucleotide bases
    # of a molecule of DNA, the amino acid residues of a protein,
    # or the atoms and chemical bonds of a small molecule.
    @property
    def elements(self):
        return str(self._elements.value)

    @elements.setter
    def elements(self, new_elements):
        self._elements.value = new_elements

    # The encoding property is REQUIRED and has a data type of URI.
    # This property MUST indicate how the elements property of a Sequence
    # MUST be formed and interpreted. For example, the elements property
    # of a Sequence with an IUPAC DNA encoding property MUST contain characters
    # that represent nucleotide bases, such as a, t, c, and g.
    # The elements property of a Sequence with a Simplified Molecular-Input
    # Line-Entry System (SMILES) encoding, on the other hand,
    # MUST contain characters that represent atoms and chemical bonds,
    # such as C, N, O, and =.
    #
    # It is RECOMMENDED that the encoding property contains a URI
    # from the table below. The terms in the table are organized
    # by the type of ComponentDefinition that typically refer to a Sequence
    # with such an encoding.
    # When the encoding of a Sequence is well described by one of the URIs
    # in the table, it MUST contain that URI.
    # | ComponentDefinition type  | Encoding       | libSBOL Symbol              | URI                                              |
    # | :------------------------ | :--------------| :-------------------------- | :----------------------------------------------- |
    # | DNA, RNA                  | IUPAC DNA, RNA | SBOL_ENCODING_IUPAC         | http://www.chem.qmul.ac.uk/iubmb/misc/naseq.html |
    # | Protein                   | IUPAC Protein  | SBOL_ENCODING_IUPAC_PROTEIN | http://www.chem.qmul.ac.uk/iupac/AminoAcid/      |
    # | Small Molecule            | SMILES         | SBOL_ENCODING_SMILES        | http://www.opensmiles.org/opensmiles.html        |
    @property
    def encoding(self):
        return self._encoding.value

    @encoding.setter
    def encoding(self, new_encoding):
        self._encoding.value = new_encoding

    def assemble(self, composite_sequence=""):
        """Calculates the complete sequence of a high-level Component
        from the sequence of its subcomponents.

        Prior to assembling the the complete sequence, you must assemble
        a template design by calling ComponentDefinition::assemble for the
        ComponentDefinition that references this Sequence.

        :param composite_sequence: Typically no value for the composite
        sequence should be specified by the user.
        This parameter is used to hold the composite sequence
        as it is passed to function calls at a higher-level of the
        recursion stack.
        :return: The complete sequence of a high-level Component.
        """
        raise NotImplementedError('Not yet implemented')

    def compile(self):
        """Synonymous with Sequence::assemble.

        Calculates the complete sequence of a high-level Component
        from the sequence of its subcomponents.
        Prior to assembling the the complete sequence, you must assemble
        a template design by calling ComponentDefinition::assemble for the
        ComponentDefinition that references this Sequence.

        :return:
        """
        self.assemble()
        return self.elements

    def __len__(self):
        """

        :return: The length of the primary sequence in the elements property.
        """
        return len(self.elements)

    def __bool__(self):
        # Ignore list semantics for boolean conversion.
        # A sequence is True if it exists.
        return True

    @deprecated(reason='Use len(sequence) instead')
    def length(self):
        """

        :return: The length of the primary sequence in the elements property.
        """
        return len(self)

    def synthesize(self, clone_id):
        """

        :param clone_id: A URI for the build, or displayId if working in
        SBOLCompliant mode.
        :return:
        """
        raise NotImplementedError("Not yet implemented")



from .toplevel import TopLevel
from .property import URIProperty
from .constants import *


class Model(TopLevel):
    def __init__(self, rdf_type=SBOL_MODEL, uri=URIRef('example'), source='',
                 language=EDAM_SBML, framework=SBO_CONTINUOUS,
                 version=VERSION_STRING):
        super().__init__(rdf_type, uri, version)
        self._source = URIProperty(self, SBOL_SOURCE,
                                   '0', '1', [], source)
        self._language = URIProperty(self, SBOL_LANGUAGE,
                                     '0', '1', [], language)
        self._framework = URIProperty(self, SBOL_FRAMEWORK,
                                      '0', '1', [], framework)

    @property
    def source(self):
        return self._source.value

    @source.setter
    def source(self, new_source):
        self._source.set(new_source)

    @property
    def language(self):
        return self._language.value

    @language.setter
    def language(self, new_language):
        self._language.set(new_language)

    @property
    def framework(self):
        return self._framework.value

    @framework.setter
    def framework(self, new_framework):
        self._framework.set(new_framework)



from .toplevel import TopLevel
from .constants import *
from rdflib import URIRef
from .property import URIProperty, LiteralProperty


class Attachment(TopLevel):

    def __init__(self, type_uri=SBOL_ATTACHMENT, uri=URIRef("example"),
                 version=VERSION_STRING, source=''):
        super().__init__(type_uri, uri, version)
        self._source = URIProperty(self, SBOL_SOURCE, '1', '1', [], source)
        self._format = LiteralProperty(self, SBOL_URI + '#format', '0', '1', [])
        self._size = LiteralProperty(self, SBOL_URI + '#size', '0', '1', [])
        self._hash = LiteralProperty(self, SBOL_URI + '#hash', '0', '1', [])

    @property
    def source(self):
        return self._source.value

    @source.setter
    def source(self, new_source):
        self._source.set(new_source)

    @property
    def format(self):
        return self._format.value

    @format.setter
    def format(self, new_format):
        self._format.set(new_format)

    @property
    def size(self):
        return self._size.value

    @size.setter
    def size(self, new_size):
        self._size.set(new_size)

    @property
    def hash(self):
        return self._hash.value

    @hash.setter
    def hash(self, new_hash):
        self._hash.set(new_hash)
